/**********************************************
***
*** uBlaze GA algorithm for the equalizer core
*** 
*** Author: Boris Spokoinyi
*** Date: 20/03/05
***
***
*** This implementation is does not talk to the core
*** It only tests if the algorithm can find the minimum
***
**********************************************/


/**************************************
***
*** CONSTANTS DECLARATION
***
**************************************/
#define POPSIZE      20
#define WEIGHTS      20
#define MUTPROB      20   //Out of 128
#define BREEDPROB   110   //Out of 128
#define MUTSIGMA      5   //Mutation amplitude=[-128..128]/2^MUTSIGMA
#define RANDNUMS    128   //Number of random numbers in ram
#define INPUTSIZE    10   //Number of bits in input sequence
/************************************/


/**************************************
*** Structure for an individual
***
*** Error is 16bits wide, overflow is taken care of in the core
*** if it overflows core will saturate the error at 0xFFFF
**************************************/
struct individual{
    signed char   weights[WEIGHTS];
    short         error;
};


/**************************************
*** VARIABLES DECLARATION
***
**************************************/
/*** precomputed random numbers:     */
signed char randomNumber[RANDNUMS]={-23,-20,-125,36,-94,68,68,90,46,-108,-87,4,52,-96,76,-9,111,-106,69,-54,-68,-124,34,-92,104,42,33,-109,109,43,83,46,72,-7,49,116,42,93,122,-55,-66,80,42,-95,66,13,-114,-113,116,28,89,-101,19,6,87,-22,-24,-8,-108,-66,-100,31,-97,-16,80,-32,36,-120,85,-107,65,72,-11,59,50,-88,72,116,90,95,74,50,-61,-124,83,31,39,48,-113,8,65,1,91,-56,-3,-104,-12,103,-35,-1,-107,-5,-55,72,49,17,53,86,-18,-65,106,54,-30,52,56,-33,-24,116,5,-2,-35,113,-3,116,74,23,101,-16};
/*** the population (a set of POPSIZE weight vectors):  */
struct individual population[POPSIZE];
/*** loop variables:
***  i,j,k: common local loop varaibles
***  jj,jjj: loop variables used in output of main
***  ii: generation counter, also used as a seed in random number lookup
***/
short i,j,k,jj,jjj,ii=0;
/*** seed:
***  currently is set to 0, in future it can be set to some random number
***  comming from outside of uBlaze, for example an input sequence
***  GA performance is nearly unaffected if seed is used in this manner
***/
short seed=0;


/**************************************
***
*** FUNCTIONS DECLARATION
***
**************************************/

/**************************************
*** FUNCTION 'absolute()'
***
*** This function is used instead of math library's abs()
*** because it's a simpler implementation of abs
***
*** Inputs: (short) number
*** Outputs: (short) absolute value of input
***
**************************************/
short absolute(short input){
    if (input>=0) return input;
    if (input<0) return -input;
}

/**************************************
*** FUNCTIONS 'getRandNum1()' and 'getRandNum2()'
***
*** The function does a table lookup of a precomputed sequence of random numbers
*** based on a seed witch is an input to the function.
*** 'getRandNum1' and 'getRandNum2' are differnt only in how they use the seed to access data
*** generally these functions will give a different random number for the same seed
***
*** Inputs: (short) a seed number
*** Outputs: (signed char) a number from a lookup table 'randomNumber[]'
*** output will be the same for a given seed
***
**************************************/
signed char getRandNum1(short seedNumber){
    return randomNumber[(seedNumber*3+ii)% RANDNUMS];
}

signed char getRandNum2(short seedNumber){
    return randomNumber[(seedNumber*2+ii) % RANDNUMS];
}

/**************************************
*** FUNCTION 'init()'
***
*** The function initialized the population, used at the beginning of main
*** weights are set to random numbers. 
*** It is very important that at the beginning the weights are as random as possible.
*** The closer the initialized weights to actiual values the faster the convergence, 
*** so it's nice to bound the weights to the most probable values
***
**************************************/
void init(){
    for (i=0;i<POPSIZE;i++){
        for (j=0;j<WEIGHTS;j++) population[i].weights[j]=getRandNum2(seed+i+j); //*** assig the weight to a random number
	//*** set initial err to some large number (not really required)
        population[i].error=10000;
    }
}

/**************************************
*** FUNCTION 'mutate()'
***
*** The function applies a random perturbation to an indivudual if it's selected.
*** Function skips elite individuals since we they are used to keep track of the best solution and we dont want to lose it.
*** First the check is performed to see if an individual shoud be mutated, this is done by looking up a random numebr
*** and looking if it falls within the -MUTPROB and MUTPROB, if yes then it's mutated. the higher MUTPROB the more chances it will happen
*** 
*** The ammount of mutation is controlled by MUTSIGMA, it basically scales the drawn random number 
*** so that the mutation ammount is not too high. Care is talken not to right shift (ie divide) a neg number
*** since wether shifted num can change sign is different for different compilers. 
*** If num is negative its first converted to positive then right-shited then sign is restored
*** 
**************************************/
void mutate(){
    for(i=2;i<POPSIZE;i++){ //skip elete one
        if (absolute(getRandNum2(seed+i))<MUTPROB){ //*** 
            //*** mutate this individual:
            for (j=0;j< WEIGHTS;j++){
                if ((getRandNum1(seed+j+i))>0)
                    population[i].weights[j]+=(getRandNum1(seed+j+i))>>MUTSIGMA;
                else
                    population[i].weights[j]+=-(absolute((getRandNum1(seed+j+i))<<1)>>(MUTSIGMA+1));

            }
        }
}

/**************************************
*** FUNCTION 'crossbreed()'
***
*** The function is used to crossbreed the individuals to produce offspring that replace the parents (but elites are kept)
*** the parents for breeding are chosen at random, in the outer loop first parent is chosen and in the innter loop second
*** parent is chosen, the resulting child is stored at the location of the first parent
*** The parents are chosen if a random number falls within the -BREEDPROB ad BREEDPROB or if it's an elite.
*** After the parents are chosen the weights are simply averaged between the parents.
*** 
**************************************/
void crossbreed(){
    for(i=2;i<POPSIZE;i++){//index is used to store offspring, we don't want to overwrite the elite
        for (j=0;j<POPSIZE;j++){
            //*** if breeding is chosen to happen or it's one of the elites and the parents are different then breed
            if ((absolute(getRandNum2(seed+i+j))<BREEDPROB  || (j==0 || j==1)) && i!=j){
                for (k=0;k<WEIGHTS;k++){ //*** do averaging of all the weights between two parents
                    if (population[i].weights[k]+population[j].weights[k]<0)
                        population[i].weights[k]=-((-(population[i].weights[k]+population[j].weights[k]))>>1); //*** div by 2 ie take the average
                    else
                        population[i].weights[k]=(population[i].weights[k]+population[j].weights[k])>>1;
                }
            }
        }
    }
}

/**************************************
*** FUNCTION 'calcErrs()'
*** 
*** The function calculates a test error surface.
*** Error is found by simply adding squares of all the weights.
*** error is scaled by 2^3=8 to avoid overflow. 
*** In real design overlow doesn't happen since equalizer will saturate error instead of overflowing.
***
**************************************/
void calcErrs(){
short sum;
    for (i=0;i<POPSIZE;i++){
        //upload all weights to the core
        //wait for error-calc-finished flag

    //for testing: simulates error response, err=sum(all weights squared)
    //min err (=0) is when all weights are zeros
        sum=0;
        sum1=0;
        for (j=0;j<WEIGHTS;j++){
            sum=sum+(population[i].weights[j]*population[i].weights[j])>>3;
            }
        }
        population[i].error=sum;
    }
}

/**************************************
*** FUNCTION 'findElite()'
*** 
*** The function does sort for the first two elements (elites)
*** this is required so that we know where the best results are (and dont mutate them without knowing)
***
**************************************/
void findElite(){
    short int minErr=population[0].error, minPos=0;
    struct individual tempInd; //for exchange of the best and the current elements
    for(i=0;i<2;i++){//look for 2 best elements

        minErr=population[i].error;
        minPos=i;
        for (j=i+1;j<POPSIZE;j++){ //find minimum
            if (population[j].error<minErr){
                minErr=population[j].error;
                minPos=j;
            }
        }

	//exchange current element with the best one:
        for (j=0;j<WEIGHTS;j++)
        tempInd.weights[j]=population[i].weights[j];
        tempInd.error=population[i].error;
        for (j=0;j<WEIGHTS;j++)
        population[i].weights[j]=population[minPos].weights[j];
        population[i].error=population[minPos].error;
        for (j=0;j<WEIGHTS;j++)
        population[minPos].weights[j]=tempInd.weights[j];
        population[minPos].error=tempInd.error;
    }
}

/**************************************
*** FUNCTION 'itterateGen()'
*** 
*** Perform all the Genetic operations in order. This is done at each generation itteration
***
**************************************/
void itterateGen(){
    crossbreed();
    calcErrs();
    findElite();
    mutate();
}

int main(void)
{
init();
for (ii=0;ii<500;ii++){
    itterateGen();
    if (ii % 10 ==0){
        xil_printf("%d,",ii);
        for (jj=0;jj<3;jj++){
            for (jjj=0;jjj<WEIGHTS;jjj++) 
            xil_printf("%d,",population[jj].weights[jjj]);
            xil_printf("er=%d \n\r",population[jj].error);
        }
        xil_printf("\n\r");
    }
}
return 0;
}