
// -----------------------------------------------------------
//  Dynamic Equalizer Core               equalizer_v1_00_b
//  Ryan Janzen & Boris Spokoinyi
//
//  Equalizer Module  - 8 Stages
//  Last updated: 2005.03.27
// -----------------------------------------------------------


module equalizer ( clk, reset_b, resetEq, sampleReady, eqDoneSample, coeffArray, signalIn, signalOut);

	input clk, reset_b, resetEq, sampleReady;
	input [63:0] coeffArray;   
	input [15:0] signalIn;
	output [31:0] signalOut;  // Size = (16bit signalIn) + (8bit coeff mult) + (5bit #coeffs=32) = 29bit
	output eqDoneSample;
	
	reg [7:0] i;   // Synthesis Counter var

	reg [127:0] chain;   // (16*nStages-1)
	reg [1:0] presentState, nextState;  // Want nextState to be optimized as non-register.
	wire multGo;
	wire multDone0;
	wire [35:0] product0, product1, product2, product3, product4, product5, product6, product7; //18x18 mults
	wire [31:0] signalOut;
	wire eqDoneSample;
	wire [15:0] const16b0;
	wire [9:0] const10b0;
	wire [7:0] const8b0;
	wire [17:0] coeff0,coeff1,coeff2,coeff3,coeff4,coeff5,coeff6,coeff7;
	wire [17:0] tap0,tap1,tap2,tap3,tap4,tap5,tap6,tap7;
	
	assign const16b0 = 16'b0000000000000000;
	assign const10b0 = 10'b0000000000;
	assign const8b0 = 8'b00000000;
	
	signExtend_8_18 coeffExt0(coeffArray[ 7: 0], coeff0);
	signExtend_8_18 coeffExt1(coeffArray[15: 8], coeff1);
	signExtend_8_18 coeffExt2(coeffArray[23:16], coeff2);
	signExtend_8_18 coeffExt3(coeffArray[31:24], coeff3);
	signExtend_8_18 coeffExt4(coeffArray[39:32], coeff4);
	signExtend_8_18 coeffExt5(coeffArray[47:40], coeff5);
	signExtend_8_18 coeffExt6(coeffArray[55:48], coeff6);
	signExtend_8_18 coeffExt7(coeffArray[63:56], coeff7);

	signExtend_16_18 tapExt0(chain[15: 0], tap0);
	signExtend_16_18 tapExt1(chain[31:16], tap1);
	signExtend_16_18 tapExt2(chain[47:32], tap2);
	signExtend_16_18 tapExt3(chain[63:48], tap3);
	signExtend_16_18 tapExt4(chain[79:64], tap4);
	signExtend_16_18 tapExt5(chain[95:80], tap5);
	signExtend_16_18 tapExt6(chain[111:96], tap6);
	signExtend_16_18 tapExt7(chain[127:112], tap7);

	// Parallel mult by coefficients
	assign multDone0=1'b1; // Xilinx hard multipliers - < 1 clock cycle?
	MULT18X18 multiplier0(product0, coeff0, tap0 );
	MULT18X18 multiplier1(product1, coeff1, tap1 );
	MULT18X18 multiplier2(product2, coeff2, tap2 );
	MULT18X18 multiplier3(product3, coeff3, tap3 );
	MULT18X18 multiplier4(product4, coeff4, tap4 );
	MULT18X18 multiplier5(product5, coeff5, tap5 );
	MULT18X18 multiplier6(product6, coeff6, tap6 );
	MULT18X18 multiplier7(product7, coeff7, tap7 );

	// Adder  *Truncates upper bits of signed sum.
	assign signalOut = product0+product1+product2+product3+product4+product5+product6+product7;   

	// Delay Line, nStages
	always @(posedge clk or negedge reset_b)
	begin
		if (reset_b==1'b0)
		begin
			chain[127:0] <= {const16b0,const16b0,const16b0,const16b0,const16b0,const16b0,const16b0,const16b0};
		end
		else if (resetEq==1'b1)
		begin
			chain[127:0] <= {const16b0,const16b0,const16b0,const16b0,const16b0,const16b0,const16b0,const16b0};
		end
		else if (presentState == 2'b01)
		begin
			chain[127:0] <= {chain[111:0],signalIn[15:0]};
		end
	end

	// State machine, sequential
	always @(posedge clk or negedge reset_b)
	begin
		if (reset_b==1'b0)
		begin
			presentState <= 2'b00;
		end
		else
		begin
			presentState <= nextState;
		end
	end

	// State machine, combinational
	always @(sampleReady or presentState or multDone0)
	begin
		if (presentState==2'b00)   // Waiting for next sample
		begin
			if (sampleReady==1'b1)
			begin
				nextState = 2'b01;
			end
			else
			begin
				nextState = 2'b00;
			end
		end
		else if (presentState==2'b01)  // Shift new sample onto chain (once) 
		begin
			nextState = 2'b10;
		end
		else if (presentState==2'b10)  // Working
		begin
			if (multDone0==1'b1)
			begin
				nextState = 2'b11;   // **Only checking first mult. See 'always' dependancy above.
			end
			else
			begin
				nextState = 2'b10;   // Keep on working
			end
		end
		else if (presentState==2'b11)  // Waiting for sampleReady assertion to be cleared.
		begin
			if (sampleReady==1'b0)
			begin
				nextState = 2'b00;     // Next sample available. shift chain, then mult. 
			end
			else
			begin
				nextState = 2'b11;    // Still waiting.
			end
		end
		else
		begin
			nextState = 2'b00;
		end
	end

	// Multipliers will start working at state '10'.
	assign multGo = (presentState == 2'b10) ? 1'b1:1'b0;
	
	assign eqDoneSample = (presentState == 2'b11) ? 1'b1:1'b0;

endmodule
