
// -----------------------------------------------------------
//  Dynamic Equalizer Core               equalizer_v1_00_b
//  Ryan Janzen & Boris Spokoinyi
//
//  Top-Level HDL for signal processor, includes Equalizer
//  Last updated: 2005.03.27
// -----------------------------------------------------------


module signalproc ( clk, reset_b, coeffArrayIn, statusIn, signalIn, 
		statusOut, statusOutIsValid, 
		dataBlockOut, dataBlockIsValid,
		errorRate, errorRateIsValid, signalOut);
		
	// -----------------------------------------------------------
	//  I/O
	// -----------------------------------------------------------
		
	input clk, reset_b;
	input [63:0] coeffArrayIn;  // Must be latched externally
	input [7:0] statusIn;
	input [15:0] signalIn;
	output [7:0] statusOut;
	output statusOutIsValid;
	output [7:0] dataBlockOut;
	output dataBlockIsValid;
	output [15:0] errorRate;
	output errorRateIsValid;
	output [31:0] signalOut;

	// -----------------------------------------------------------
	//  Wires and Regs
	// -----------------------------------------------------------

	reg [7:0] counterSampleBlock;
	reg [1:0] presentState, nextState;  // Want nextState optimized as wire.
	reg sampleReadyPrev;  
	wire [31:0] signalOut;
	wire binaryOut;
	wire [15:0] errorRate;
	wire eqDoneSample, resetEq, sampleReady, signalproc_goCmd, sampleTrigger;
	// wire sampleTrigger, byteTrigger, blockTrigger;
	reg [7:0] byteOutShiftReg;
	wire [15:0] const16b0;

	assign const16b0 = 16'b0000000000000000;

	// -----------------------------------------------------------
	//  Control I/O
	// -----------------------------------------------------------

	assign signalproc_goCmd = statusIn[2];
	assign sampleReady = statusIn[0];
	
	assign statusOut = {presentState, counterSampleBlock[2:0], dataBlockIsValid, 
						errorRateIsValid, eqDoneSample};
	assign statusOutIsValid = 1'b1;
	assign dataBlockOut = byteOutShiftReg;
	assign dataBlockIsValid = (presentState==2'b10) ? 1'b1:1'b0;


	// -----------------------------------------------------------
	//  Signal Processing
	// -----------------------------------------------------------

	errorChecker ec1( clk, reset_b, sampleTrigger, counterSampleBlock, signalOut, errorRate, errorRateIsValid);
	equalizer eq1( clk, reset_b, resetEq, sampleReady, eqDoneSample, coeffArrayIn, signalIn, signalOut);
	
	// Threshold Detector, Analog->Binary
	assign binaryOut = (signalOut[31]==1'b0) ? 1'b1:1'b0;  // Positive->1, Neg->0   **Set index to highest bit.


	// -----------------------------------------------------------
	//  Data Protocol - Counting & Sequencing
	// -----------------------------------------------------------

	// New sample indicator - Delay latch
	always @(posedge clk or negedge reset_b)	
	begin
		if (reset_b==1'b0)	
		begin
			sampleReadyPrev <= 1'b0;
  		end
		else
		begin
  			sampleReadyPrev <= sampleReady;
		end
	end

	assign sampleTrigger = (sampleReady==1'b1 && sampleReadyPrev ==1'b0 && presentState==2'b01) ? 1'b1:1'b0;
	

	// Counts which sample we're at within "data block"
	always @(posedge clk or negedge reset_b)
	begin	
		if (reset_b==1'b0)	
		begin		
			counterSampleBlock <= 8'b00000000;
  		end	
		else if (presentState==2'b00)	
		begin		
			counterSampleBlock <= 8'b00000000;
  		end	
  		else if (sampleTrigger==1'b1 && presentState==2'b01)
  		begin		
  			counterSampleBlock <= counterSampleBlock + 1;
		end
	end	
	
	// Data Out:  Serial -> 8bit parallel converter  (Shift reg)	
	always @(posedge clk or negedge reset_b)
	begin
		if (reset_b==1'b0)	
		begin
			byteOutShiftReg <= 8'b00000000;
  		end
  		else if (sampleTrigger==1'b1 && errorRateIsValid)  // If not in PRS region.
  		begin
  			byteOutShiftReg <= {binaryOut,byteOutShiftReg[7:1]};
		end
	end


	// -----------------------------------------------------------
	//  State Machine (signalproc)
	// -----------------------------------------------------------

	// State machine, sequential
	always @(posedge clk or negedge reset_b)
	begin
		if (reset_b==1'b0)
		begin
			presentState <= 2'b00;
		end
		else
		begin
			presentState <= nextState;
		end
	end

	// State machine, combinational
	always @(presentState or signalproc_goCmd or eqDoneSample or counterSampleBlock)
	begin
		if (presentState==2'b00)   // Reset EQ signal chain.
		begin
			if (signalproc_goCmd==1'b1)
			begin
				nextState = 2'b01;
			end
			else
			begin
				nextState = 2'b00;
			end
		end
		else if (presentState==2'b01)  // Working.
			if (counterSampleBlock == 8'b00010100 && eqDoneSample==1'b1)
			//								  ^Latency
			begin
				nextState = 2'b10;   // Last sample in the block is finished.
			end
			else
			begin
				nextState = 2'b01;   // Keep on working
			end
		else if (presentState==2'b10)   // Done
		begin
			if (signalproc_goCmd==1'b0)  // Acknowledgement that data was read.
			begin
				nextState = 2'b00;   // Go back to start; clear reg's.
			end
			else
			begin
				nextState = 2'b10;   // Wait for external to read output
			end
		end
		else
		begin
			nextState = 2'b00;
		end
	end

	assign resetEq = (presentState==2'b00) ? 1'b1:1'b0;
	
endmodule